/* 
 * Copyright 1994 Alastair Gourlay
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice appear in all copies and that both that 
 * copyright notice and this permission notice appear in supporting 
 * documentation.
 */

/*
 *
 * Bezel.c - XmpBezel widget
 *
 * XmpGeometry subclass. Simple Bezel widget.
 *
 */


#include <stdio.h>
#include <Xm/DrawP.h>	/* csk: added for prototype of _XmDrawShadows() */
#include "BezelP.h"

static void Initialize(Widget, Widget, ArgList, Cardinal*);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static void Size(Widget,Dimension*,Dimension*);
static void Resize(Widget);
static void Redisplay(Widget,XEvent*,Region);

static XtResource resources[] = 
{
    {
	XmNbezelThickness,
	XmCBezelThickness,
	XmRHorizontalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpBezelRec, bezel.bezel_thickness),
	XtRImmediate,
	(XtPointer) 10
    },
    {
	XmNmarginThickness,
	XmCMarginThickness,
	XmRHorizontalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpBezelRec, bezel.margin_thickness),
	XtRImmediate,
	(XtPointer) 10
    },
};


static XmSyntheticResource syn_resources[] =
{
    { 
	XmNbezelThickness,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpBezelRec, bezel.bezel_thickness),
	_XmFromHorizontalPixels,
	_XmToHorizontalPixels 
    },
    { 
	XmNmarginThickness,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpBezelRec, bezel.margin_thickness),
	_XmFromHorizontalPixels,
	_XmToHorizontalPixels 
    },
};



/****************************************************************
 *
 * XmpBezel class record
 *
 ****************************************************************/

externaldef(xmbezelclassrec) XmpBezelClassRec xmpBezelClassRec = 
{
  {                                           /* core_class            */
    (WidgetClass)&xmpGeometryClassRec,        /* superclass            */
    "XmpBezel",                               /* class_name            */
    sizeof(XmpBezelRec),                      /* widget_size           */
    NULL,                                     /* class_initialize      */
    NULL,                                     /* class_part_initialize */
    FALSE,                                    /* class_inited          */
    Initialize,                               /* initialize            */
    NULL,                                     /* initialize_hook       */
    XtInheritRealize,                         /* realize               */
    NULL,                                     /* actions               */
    0,                                        /* num_actions           */
    resources,                                /* resources             */
    XtNumber(resources),                      /* num_resources         */
    NULLQUARK,                                /* xrm_class             */
    TRUE,                                     /* compress_motion       */
    XtExposeCompressMaximal,                  /* compress_exposure     */
    TRUE,                                     /* compress_enterleave   */
    FALSE,                                    /* visible_interest      */
    NULL,                                     /* destroy               */
    Resize,                                   /* resize                */
    Redisplay,                                /* expose                */
    SetValues,                                /* set_values            */
    NULL,                                     /* set_values_hook       */
    XtInheritSetValuesAlmost,                 /* set_values_almost     */
    NULL,                                     /* get_values_hook       */
    NULL,                                     /* accept_focus          */
    XtVersion,                                /* version               */
    NULL,                                     /* callback_private      */
    XtInheritTranslations,                    /* tm_table              */
    XtInheritQueryGeometry,                   /* query_geometry        */
    NULL,                                     /* display_accelerator   */
    NULL,                                     /* extension             */
  },

  {                                           /* composite_class       */
    XtInheritGeometryManager,                 /* geometry_manager      */
    XtInheritChangeManaged,                   /* change_managed        */
    XtInheritInsertChild,                     /* insert_child          */
    XtInheritDeleteChild,                     /* delete_child          */
    NULL,                                     /* extension             */
  },

  {                                           /* constraint_class      */
    NULL,                                     /* resources             */   
    0,                                        /* num_resources         */   
    sizeof(XmpBezelConstraintRec),            /* constraint_size       */
    NULL,                                     /* constraint_initialize */
    NULL,                                     /* destroy               */   
    NULL,                                     /* set_values            */   
    NULL,                                     /* extension             */
  },

  {                                           /* manager class         */
    XtInheritTranslations,                    /* translations          */
    syn_resources,                            /* syn_resources         */
    XtNumber(syn_resources),                  /* num_syn_resources     */
    NULL,                                     /* syn_constraint_resources */
    0,                                        /* num_syn_constraint_resources */
    XmInheritParentProcess,                   /* parent_process        */
    NULL,                                     /* extension             */    
  },

  {                                           /* geometry class        */
    XmpInheritBitGravity,                     /* bit_gravity           */
    XmpInheritInitializePostHook,             /* initialize_post_hook  */
    XmpInheritSetValuesPostHook,              /* set_values_post_hook  */
    XmpInheritConstraintInitializePostHook,   /* constraint_initialize_post_hook*/
    XmpInheritConstraintSetValuesPostHook,    /* constraint_set_values_post_hook*/
    Size,                                     /* size                  */
    NULL,                                     /* extension             */    
  },

  {                                           /* bezel class           */
    NULL,                                     /* extension             */    
  },        
};

externaldef(xmpbezelwidgetclass) WidgetClass xmpBezelWidgetClass =
			                        (WidgetClass) &xmpBezelClassRec;


/************************************************************************
 *
 *  Initialize
 *
 ************************************************************************/
static void
Initialize (Widget request_w, Widget new_w, ArgList args, Cardinal* num_args)
{
    XmpGeometryInitialize(xmpBezelWidgetClass, request_w, new_w, args, num_args);
}



/************************************************************************
 *
 *  Set Values
 *
 ************************************************************************/
static Boolean
SetValues (Widget old_w, Widget request_w, Widget new_w, ArgList args,
							Cardinal* num_args)
{
    XmpBezelWidget cw = (XmpBezelWidget)old_w;
    XmpBezelWidget nw = (XmpBezelWidget)new_w;
    Boolean redisplay = False;

    if (BezelThickness(nw) != BezelThickness(cw) ||
	MarginThickness(nw) != MarginThickness(cw)) {

	Reconfigure(nw) = True;
    }

    redisplay |= XmpGeometrySetValues(xmpBezelWidgetClass,
					old_w, request_w, new_w, args, num_args);

    return (redisplay);
}



/************************************************************************
 *
 *  Size
 *
 ************************************************************************/
static void
Size (Widget w, Dimension* bezelWidth, Dimension* bezelHeight)
{
    XmpBezelWidget gw = (XmpBezelWidget)w;
    Dimension maxWidth = 1;
    Dimension maxHeight = 1;
    int i;

    for (i=0; i < NumChildren(gw); i++) {
	Widget ic = Children(gw)[i];
	Dimension cw, ch, cb;

	if (!XtIsManaged(ic)) continue;

	XmpPreferredGeometry(ic,&cw, &ch);
	cb = BorderWidth(ic);

	maxWidth = Max(maxWidth, cw+2*cb);
	maxHeight = Max(maxHeight, ch+2*cb);
    }

    *bezelWidth = maxWidth + 2 * (BezelThickness(gw) + MarginThickness(gw));
    *bezelHeight = maxHeight + 2 * (BezelThickness(gw) + MarginThickness(gw));
}



/************************************************************************
 *
 *  Resize 
 *
 ************************************************************************/
static void
Resize (Widget w)
{
    XmpBezelWidget gw = (XmpBezelWidget)w;
    Dimension borderThickness = BezelThickness(gw) + MarginThickness(gw);
    Dimension availWidth = XmpSubtract(Width(gw), 2u * borderThickness, 1);
    Dimension availHeight = XmpSubtract(Height(gw), 2u * borderThickness, 1);
    int i;

    for (i = 0; i < NumChildren(gw); i++) {
	Widget ic = Children(gw)[i];
	Dimension cw, ch, cb;

	if (!XtIsManaged(ic)) continue;

	cb = BorderWidth(ic);
	cw = XmpSubtract(availWidth, 2u*cb, 1);
	ch = XmpSubtract(availHeight, 2u*cb, 1);

	XmpSetGeometry(ic,borderThickness,borderThickness,cw,ch,cb);
    }
}



/************************************************************************
 *
 *  Redisplay
 *
 ************************************************************************/
static void
Redisplay (Widget w, XEvent* event, Region region)
{
    XmpBezelWidget gw = (XmpBezelWidget)w;

    if (BezelThickness(gw) >= 1)
	_XmDrawShadows(XtDisplay(gw), XtWindow(gw),
		TopShadowGC(gw), BottomShadowGC(gw),
		0, 0,
		Width(gw), Height(gw), 1,
		XmSHADOW_OUT);

    if (BezelThickness(gw) >= 2)
	_XmDrawShadows(XtDisplay(gw), XtWindow(gw),
		TopShadowGC(gw), BottomShadowGC(gw),
		BezelThickness(gw)-1, BezelThickness(gw)-1,
		XmpSubtract(Width(gw),2u*BezelThickness(gw)-2, 1),
		XmpSubtract(Height(gw),2u*BezelThickness(gw)-2, 1),
		1,
		XmSHADOW_IN);

    XmpGeometryRedisplay(w, event, region);
}



/************************************************************************
 *
 *  XmpCreateBezel
 *
 ************************************************************************/
Widget
XmpCreateBezel(Widget parent, char* name, Arg* args, Cardinal num_args)
{
    return(XtCreateWidget(name, xmpBezelWidgetClass, parent, args, num_args));
}

