/* 
 * Copyright 1994 Alastair Gourlay
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice appear in all copies and that both that 
 * copyright notice and this permission notice appear in supporting 
 * documentation.
 */

/*
 *
 * Caption.c - XmpCaption widget
 *
 * XmpGeometry subclass. Simple Caption widget.
 * 
 */


#include <stdio.h>
#include "CaptionP.h"

static void ClassInitialize(void);
static void Initialize(Widget, Widget, ArgList, Cardinal*);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static void ConstraintInitialize(Widget,Widget,ArgList,Cardinal*);
static Boolean ConstraintSetValues(Widget,Widget,Widget,ArgList,Cardinal*);
static void Size(Widget,Dimension*,Dimension*);
static void Resize(Widget);
static Widget GetNextChild(Widget,int*);

static XtResource resources[] = 
{
    {
        XmNstrechControl,
        XmCStrechControl,
        XmRBoolean,
        sizeof (Boolean),
	XtOffsetOf(struct _XmpCaptionRec, caption.strech_control),
	XtRImmediate,
	(XtPointer)True
    },
    {
	XmNmarginWidth,
	XmCMarginWidth,
	XmRHorizontalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpCaptionRec, caption.margin_width),
	XtRImmediate,
	(XtPointer)2
    },
    {
	XmNmarginHeight,
	XmCMarginHeight,
	XmRVerticalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpCaptionRec, caption.margin_height),
	XtRImmediate,
	(XtPointer)2
    },
    {
	XmNhorizontalSpacing,
	XmCSpacing,
	XmRHorizontalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpCaptionRec, caption.horizontal_spacing),
	XtRImmediate,
	(XtPointer)2
    },
    {
	XmNverticalSpacing,
	XmCSpacing,
	XmRVerticalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpCaptionRec, caption.vertical_spacing),
	XtRImmediate,
	(XtPointer)8
    },
};

static XmSyntheticResource syn_resources[] =
{
    { 
	XmNmarginWidth,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpCaptionRec, caption.margin_width),
	_XmFromHorizontalPixels,
	_XmToHorizontalPixels 
    },
    { 
	XmNmarginHeight,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpCaptionRec, caption.margin_height),
	_XmFromVerticalPixels,
	_XmToVerticalPixels 
    },
    { 
	XmNhorizontalSpacing,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpCaptionRec, caption.horizontal_spacing),
	_XmFromHorizontalPixels,
	_XmToHorizontalPixels 
    },
    { 
	XmNverticalSpacing,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpCaptionRec, caption.vertical_spacing),
	_XmFromVerticalPixels,
	_XmToVerticalPixels 
    },
};

static XtResource constraints[] =
{
    {
	XmNcaptionJustify,
	XmCCaptionJustify,
	XmRCaptionJustify,
	sizeof (unsigned char),
	XtOffsetOf(struct _XmpCaptionConstraintRec, caption.caption_justify),
	XtRImmediate,
	(XtPointer)XmJUSTIFY_LEFT
    },
    {
	XmNcaptionAlignment,
	XmCCaptionAlignment,
	XmRCaptionAlignment,
	sizeof (unsigned char),
	XtOffsetOf(struct _XmpCaptionConstraintRec, caption.caption_alignment),
	XtRImmediate,
	(XtPointer)XmCAPTION_CENTER
    },
    {
	XmNstretchHorizontal,
	XmCStretchHorizontal,
	XmRBoolean,
	sizeof (Boolean),
	XtOffsetOf(struct _XmpCaptionConstraintRec, caption.stretch_horizontal),
	XtRImmediate,
	(XtPointer)False
    },
    {
	XmNstretchVertical,
	XmCStretchVertical,
	XmRBoolean,
	sizeof (Boolean),
	XtOffsetOf(struct _XmpCaptionConstraintRec, caption.stretch_vertical),
	XtRImmediate,
	(XtPointer)False
    },
};



/****************************************************************
 *
 * XmpCaption class record
 *
 ****************************************************************/

externaldef(xmcaptionclassrec) XmpCaptionClassRec xmpCaptionClassRec = 
{
  {                                           /* core_class            */
    (WidgetClass)&xmpGeometryClassRec,        /* superclass            */
    "XmpCaption",                             /* class_name            */
    sizeof(XmpCaptionRec),                    /* widget_size           */
    ClassInitialize,                          /* class_initialize      */
    NULL,                                     /* class_part_initialize */
    FALSE,                                    /* class_inited          */
    Initialize,                               /* initialize            */
    NULL,                                     /* initialize_hook       */
    XtInheritRealize,                         /* realize               */
    NULL,                                     /* actions               */
    0,                                        /* num_actions           */
    resources,                                /* resources             */
    XtNumber(resources),                      /* num_resources         */
    NULLQUARK,                                /* xrm_class             */
    TRUE,                                     /* compress_motion       */
    XtExposeCompressMaximal,                  /* compress_exposure     */
    TRUE,                                     /* compress_enterleave   */
    FALSE,                                    /* visible_interest      */
    NULL,                                     /* destroy               */
    Resize,                                   /* resize                */
    XtInheritExpose,                          /* expose                */
    SetValues,                                /* set_values            */
    NULL,                                     /* set_values_hook       */
    XtInheritSetValuesAlmost,                 /* set_values_almost     */
    NULL,                                     /* get_values_hook       */
    NULL,                                     /* accept_focus          */
    XtVersion,                                /* version               */
    NULL,                                     /* callback_private      */
    XtInheritTranslations,                    /* tm_table              */
    XtInheritQueryGeometry,                   /* query_geometry        */
    NULL,                                     /* display_accelerator   */
    NULL,                                     /* extension             */
  },

  {                                           /* composite_class       */
    XtInheritGeometryManager,                 /* geometry_manager      */
    XtInheritChangeManaged,                   /* change_managed        */
    XtInheritInsertChild,                     /* insert_child          */
    XtInheritDeleteChild,                     /* delete_child          */
    NULL,                                     /* extension             */
  },

  {                                           /* constraint_class      */
    constraints,                              /* resources             */   
    XtNumber(constraints),                    /* num_resources         */   
    sizeof(XmpCaptionConstraintRec),          /* constraint_size       */
    ConstraintInitialize,                     /* initialize            */   
    NULL,                                     /* destroy               */   
    ConstraintSetValues,                      /* set_values            */   
    NULL,                                     /* extension             */
  },

  {                                           /* manager class         */
    XtInheritTranslations,                    /* translations          */
    syn_resources,                            /* syn_resources         */
    XtNumber(syn_resources),                  /* num_syn_resources     */
    NULL,                                     /* syn_constraint_resources */
    0,                                        /* num_syn_constraint_resources */
    XmInheritParentProcess,                   /* parent_process        */
    NULL,                                     /* extension             */    
  },

  {                                           /* geometry class        */
    XmpInheritBitGravity,                     /* bit_gravity           */
    XmpInheritInitializePostHook,             /* initialize_post_hook  */
    XmpInheritSetValuesPostHook,              /* set_values_post_hook  */
    XmpInheritConstraintInitializePostHook,   /* constraint_initialize_post_hook*/
    XmpInheritConstraintSetValuesPostHook,    /* constraint_set_values_post_hook*/
    Size,                                     /* size                  */
    NULL,                                     /* extension             */    
  },

  {                                           /* caption class         */
    0,                                        /* caption_alignment_id  */
    0,                                        /* caption_justify_id    */
    NULL,                                     /* extension             */    
  },        
};

externaldef(xmpcaptionwidgetclass) WidgetClass xmpCaptionWidgetClass =
			                        (WidgetClass) &xmpCaptionClassRec;

static String CaptionAlignmentNames[] = {
	"caption_top",
	"caption_center",
	"caption_bottom"
};
static String JustifyNames[] = {
        "justify_left",
        "justify_right"
};

/************************************************************************
 *
 *  ClassInitialize
 *
 ************************************************************************/
static void
ClassInitialize(void)
{
    CaptionAlignmentId = XmRepTypeRegister(XmRCaptionAlignment,
					CaptionAlignmentNames, NULL,
					XtNumber(CaptionAlignmentNames));
    CaptionJustifyId = XmRepTypeRegister(XmRCaptionJustify,
					JustifyNames, NULL,
					XtNumber(JustifyNames));
}



/************************************************************************
 *
 *  Initialize
 *
 ************************************************************************/
static void
Initialize (Widget request_w, Widget new_w, ArgList args, Cardinal* num_args)
{
    XmpGeometryInitialize(xmpCaptionWidgetClass,request_w,new_w,args,num_args);
}



/************************************************************************
 *
 *  Set Values
 *
 ************************************************************************/
static Boolean
SetValues (Widget old_w, Widget request_w, Widget new_w, ArgList args,
							Cardinal* num_args)
{
    XmpCaptionWidget cw = (XmpCaptionWidget)old_w;
    XmpCaptionWidget nw = (XmpCaptionWidget)new_w;
    Boolean redisplay = False;

    if (MarginWidth(nw) != MarginWidth(cw) ||
	MarginHeight(nw) != MarginHeight(cw) ||
	HorizontalSpacing(nw) != HorizontalSpacing(cw) ||
	VerticalSpacing(nw) != VerticalSpacing(cw)) {

	Reconfigure(nw) = True;
    }

    redisplay |= XmpGeometrySetValues(xmpCaptionWidgetClass,
				old_w, request_w, new_w, args, num_args);

    return (redisplay);
}



/************************************************************************
 *
 *  ConstraintInitialize
 *
 ************************************************************************/

static void ConstraintInitialize (
   Widget rw,
   Widget nw,
   ArgList args,
   Cardinal * num_args)
{
   if (!XtIsRectObj (nw))
      return;

   if (!XmRepTypeValidValue(CaptionAlignmentId,CaptionAlignment(nw),(Widget)nw))
      CaptionAlignment(nw) = XmCAPTION_CENTER;

   if (!XmRepTypeValidValue(CaptionJustifyId,CaptionJustify(nw),(Widget)nw))
      CaptionJustify(nw) = XmJUSTIFY_LEFT;

    XmpGeometryConstraintInitialize(xmpCaptionWidgetClass,rw,nw,args,num_args);
}



/************************************************************************
 *
 *  ConstraintSetValues
 *
 ************************************************************************/

static
Boolean
ConstraintSetValues (
   Widget cw,
   Widget rw, 
   Widget nw,
   ArgList args,
   Cardinal* num_args)
{
   XmpCaptionWidget gw = (XmpCaptionWidget)XtParent(nw);
   Boolean redisplay = False;

   if (!XtIsRectObj (nw)) return (False);

   if (CaptionAlignment(nw) != CaptionAlignment(cw))
      if (!XmRepTypeValidValue(CaptionAlignmentId, CaptionAlignment(nw), nw))
         CaptionAlignment(nw) = CaptionAlignment(cw);

    if (XtIsManaged(nw) && (
	StretchHorizontal(nw) != StretchHorizontal(cw) ||
	StretchVertical(nw) != StretchVertical(cw) ||
	CaptionAlignment(nw) != CaptionAlignment(cw)
	)) {

	ConstraintReconfigure(gw) = True;
    }

    redisplay |= XmpGeometryConstraintSetValues(xmpCaptionWidgetClass,
						cw, rw, nw, args, num_args);

    return(redisplay);
}



/************************************************************************
 *
 *  Size
 *
 ************************************************************************/
static
void
Size (
   Widget w,
   Dimension* captionWidth,
   Dimension* captionHeight )
{
    XmpCaptionWidget gw = (XmpCaptionWidget)w;
    Dimension maxWidth = 1;
    Dimension maxHeight = 1;
    Dimension minHeight = 0;
    Dimension trueMinHeight = 0;
    Dimension maxCaptionWidth = 1;
    Dimension maxControlWidth = 1;
    Dimension totalStretchPrefHeight = 0;
    Dimension lcw, lch, lcb;
    Dimension rcw, rch, rcb;
    int i;

    /* Calculate the ideal size. If application set geometry, use that. */

    i = 0;
    while(i < NumChildren(gw)) {
	Widget lw,rw;

        /* Find next managed child, it's the label */

	if ((lw = GetNextChild((Widget)gw, &i)) == NULL)
	    break;
 
	/* Well, unless it's a separator */

	if (XmIsSeparator(lw) || XmIsSeparatorGadget(lw)) {
	    Dimension leftHeight;

	    XmpPreferredGeometry(lw, NULL, &lch);
	    leftHeight = lch + 2 * BorderWidth(lw);
	    maxHeight += leftHeight;
	    minHeight += leftHeight;
	    trueMinHeight += leftHeight;
	    i += 1;
	    if (i < NumChildren(gw)) {
		maxHeight += VerticalSpacing(gw);
		minHeight += VerticalSpacing(gw);
		trueMinHeight += VerticalSpacing(gw);
	    }
	    continue;
	}

	i += 1;

        XmpPreferredGeometry(lw, &lcw, &lch);
	lcb = BorderWidth(lw);

        /* Find next managed child, it's the control */

	if ((rw = GetNextChild((Widget)gw, &i)) == NULL)
	    break;

        XmpPreferredGeometry(rw, &rcw, &rch);
	rcb = BorderWidth(rw);

	maxCaptionWidth = Max(lcw+2u*lcb,maxCaptionWidth);
	maxControlWidth = Max(rcw+2u*rcb,maxControlWidth);
	maxHeight += Max(lch+2u*lcb,rch+2u*rcb);
	if (StretchVertical(rw)==True) {
	    minHeight += lch+2*lcb;
	    totalStretchPrefHeight += rch + 2*rcb;
	}
	else {
	    minHeight += Max(lch+2u*lcb,rch+2u*rcb);
	    trueMinHeight += Max(lch+2u*lcb,rch+2u*rcb);
	}

	i += 1;
	if (i < NumChildren(gw)) {
	    maxHeight += VerticalSpacing(gw);
	    trueMinHeight += VerticalSpacing(gw);
	}
    }

    maxWidth = maxCaptionWidth + HorizontalSpacing(gw) + maxControlWidth;

    /* Cache information needed in Resize */

    MaximumCaptionWidth(gw) = maxCaptionWidth;
    MinimumUsefulHeight(gw) = minHeight + 2 * MarginHeight(gw);
    TrueMinimumHeight(gw) = trueMinHeight + 2 * MarginHeight(gw);
    TotalStretchPrefHeight(gw) = totalStretchPrefHeight;

    *captionWidth = maxWidth + 2 * MarginWidth(gw);
    *captionHeight = maxHeight + 2 * MarginHeight(gw);
}



/************************************************************************
 *
 *  Resize 
 *
 ************************************************************************/
static
void
Resize (
   Widget w )
{
    XmpCaptionWidget gw = (XmpCaptionWidget)w;
    Dimension mw = MarginWidth(gw);
    Dimension mh = MarginHeight(gw);
    Position py;
    int i;

    py = mh;
 
    i = 0;
    while(i < NumChildren(gw)) {
        Widget lw;
        Widget rw;
	Position lcx, lcy, rcx, rcy;
	Dimension lcw, lch, lcb, rcw, rch, rcb;
	Dimension usedWidth;
	Dimension availableWidth;
	Dimension prefWidth;
	Dimension prefHeight;
	Dimension rowHeight;
 
        /* Find next managed child, it's the label */
 
	if ((lw = GetNextChild((Widget)gw, &i)) == NULL)
	    break;
 
	/* Well, unless it's the separator */

        if (XmIsSeparator(lw) || XmIsSeparatorGadget(lw)) {
	    lcw = XmpSubtract(Width(gw), 2u*BorderWidth(lw), 1);
	    XmpSetGeometry(lw, 0, py, lcw, Height(lw), BorderWidth(lw));
	    py += Height(lw) + 2 * BorderWidth(lw) + VerticalSpacing(gw);
            i += 1;
            continue;
        }
 
        i += 1;

        /* Find next managed child, it's the control */

	if ((rw = GetNextChild((Widget)gw, &i)) == NULL)
	    break;

	XmpPreferredGeometry(rw, &prefWidth, &prefHeight);

	lcw = Width(lw);
	lch = Height(lw);
	lcb = BorderWidth(lw);
	rcw = Width(rw);
	rcb = BorderWidth(rw);

	if (StretchVertical(rw)==True) {
	    if (Height(gw) > MinimumUsefulHeight(gw)) {
		Dimension space = Height(gw) - TrueMinimumHeight(gw);
		Dimension boxHeight = (prefHeight + 2u*rcb) * space /
						TotalStretchPrefHeight(gw);
		rch = XmpSubtract(boxHeight, 2u*rcb, 1);
	    }
	    else {
		rch = XmpSubtract(lch+2u*lcb, 2u*rcb, 1);
	    }
	}
	else {
	   rch = prefHeight;
	}
	rowHeight = Max(lch+2u*lcb, rch+2*rcb);

        switch(CaptionJustify(lw))
        {
           case XmJUSTIFY_RIGHT:
	      lcx = mw + MaximumCaptionWidth(gw) - (lcw+2*lcb);
              break;

           case XmJUSTIFY_LEFT:
           default :
              lcx = mw;
              break;
        }

	switch(CaptionAlignment(lw)) {
	    case XmCAPTION_TOP:
		lcy = py;
		break;
	    case XmCAPTION_CENTER:
	    default:
		lcy = py + (int)(rowHeight - lch - 2*lcb)/2;
		break;
	    case XmCAPTION_BOTTOM:
		lcy = py + rowHeight - lch - 2*lcb;
		break;
	}

	rcx = mw + MaximumCaptionWidth(gw) + HorizontalSpacing(gw);
	rcy = py + (int)(rowHeight - rch -2*rcb)/2;

	usedWidth = 2 * mw + MaximumCaptionWidth(gw) + HorizontalSpacing(gw);
	availableWidth = XmpSubtract(Width(gw), usedWidth, 1u);

        if (StrechControl(gw))
            prefWidth = availableWidth;
	if (availableWidth < prefWidth+2u*rcb ||
	    StretchHorizontal(rw)==True)
	    rcw = XmpSubtract(availableWidth,2u*rcb,1);
	else
	    rcw = prefWidth;

	XmpSetGeometry(lw,lcx,lcy,lcw,lch,lcb);
	XmpSetGeometry(rw,rcx,rcy,rcw,rch,rcb);

	py += rowHeight + VerticalSpacing(gw);
	i += 1;
    }
}



/* Get next interesting child in child list, starting at [i] */

static
Widget
GetNextChild (
   Widget w,
   int* i )
{
   XmpCaptionWidget gw = (XmpCaptionWidget)w;
   Widget child = NULL;

   while (*i < NumChildren(gw)) {
      if (XtIsManaged(Children(gw)[*i])) {
         child = Children(gw)[*i];
         break;
      }
      *i += 1;
   }

   return child;
}


/************************************************************************
 *
 *  XmpCreateCaption
 *
 ************************************************************************/
Widget
XmpCreateCaption(Widget parent, char* name, Arg* args, Cardinal num_args)
{
    return(XtCreateWidget(name, xmpCaptionWidgetClass, parent, args, num_args));
}

