/*
 *
 * Template.c - XmpTemplate widget
 *
 * XmpGeometry subclass. Template widget.
 *
 */


#include <stdio.h>
#include "TemplateP.h"

static void Initialize(Widget, Widget, ArgList, Cardinal*);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static void ConstraintInitialize(Widget, Widget, ArgList, Cardinal*);
static Boolean ConstraintSetValues(Widget,Widget,Widget,ArgList,Cardinal*);
static void Size(Widget,Dimension*,Dimension*);
static void Resize(Widget);

static XtResource resources[] = 
{
    {
	XmNtemplateResource,
	XmCTemplateResource,
	XmRInt,
	sizeof (int),
	XtOffsetOf(struct _XmpTemplateRec, template.template_resource),
	XtRImmediate,
	(XtPointer)0
    },
};


static XtResource constraints[] =
{
    {
	XmNtemplateConstraintResource,
	XmCTemplateConstraintResource,
	XmRInt,
	sizeof (int),
	XtOffsetOf(struct _XmpTemplateConstraintRec,
					template.template_constraint_resource),
	XmRImmediate,
	(XtPointer)0
    },
};


/****************************************************************
 *
 * XmpTemplate class record
 *
 ****************************************************************/

externaldef(xmtemplateclassrec) XmpTemplateClassRec xmpTemplateClassRec = 
{
  {                                           /* core_class            */
    (WidgetClass)&xmpGeometryClassRec,        /* superclass            */
    "XmpTemplate",                            /* class_name            */
    sizeof(XmpTemplateRec),                   /* widget_size           */
    NULL,                                     /* class_initialize      */
    NULL,                                     /* class_part_initialize */
    FALSE,                                    /* class_inited          */
    Initialize,                               /* initialize            */
    NULL,                                     /* initialize_hook       */
    XtInheritRealize,                         /* realize               */
    NULL,                                     /* actions               */
    0,                                        /* num_actions           */
    resources,                                /* resources             */
    XtNumber(resources),                      /* num_resources         */
    NULLQUARK,                                /* xrm_class             */
    TRUE,                                     /* compress_motion       */
    XtExposeCompressMaximal,                  /* compress_exposure     */
    TRUE,                                     /* compress_enterleave   */
    FALSE,                                    /* visible_interest      */
    NULL,                                     /* destroy               */
    Resize,                                   /* resize                */
    XtInheritExpose,                          /* expose                */
    SetValues,                                /* set_values            */
    NULL,                                     /* set_values_hook       */
    XtInheritSetValuesAlmost,                 /* set_values_almost     */
    NULL,                                     /* get_values_hook       */
    NULL,                                     /* accept_focus          */
    XtVersion,                                /* version               */
    NULL,                                     /* callback_private      */
    XtInheritTranslations,                    /* tm_table              */
    XtInheritQueryGeometry,                   /* query_geometry        */
    NULL,                                     /* display_accelerator   */
    NULL,                                     /* extension             */
  },

  {                                           /* composite_class       */
    XtInheritGeometryManager,                 /* geometry_manager      */
    XtInheritChangeManaged,                   /* change_managed        */
    XtInheritInsertChild,                     /* insert_child          */
    XtInheritDeleteChild,                     /* delete_child          */
    NULL,                                     /* extension             */
  },

  {                                           /* constraint_class      */
    constraints,                              /* resources             */   
    XtNumber(constraints),                    /* num_resources         */   
    sizeof(XmpTemplateConstraintRec),             /* constraint_size       */
    ConstraintInitialize,                     /* constraint_initialize */
    NULL,                                     /* destroy               */   
    ConstraintSetValues,                      /* set_values            */   
    NULL,                                     /* extension             */
  },

  {                                           /* manager class         */
    XtInheritTranslations,                    /* translations          */
    NULL,                                     /* syn_resources         */
    0,                                        /* num_syn_resources     */
    NULL,                                     /* syn_constraint_resources */
    0,                                        /* num_syn_constraint_resources */
    XmInheritParentProcess,                   /* parent_process        */
    NULL,                                     /* extension             */    
  },

  {                                           /* geometry class        */
    XmpInheritBitGravity,                     /* bit_gravity           */
    XmpInheritInitializePostHook,             /* initialize_post_hook  */
    XmpInheritSetValuesPostHook,              /* set_values_post_hook  */
    XmpInheritConstraintInitializePostHook,   /* constraint_initialize_post_hook*/
    XmpInheritConstraintSetValuesPostHook,    /* constraint_set_values_post_hook*/
    Size,                                     /* size                  */
    NULL,                                     /* extension             */    
  },

  {                                           /* template class            */
    NULL,                                     /* extension             */    
  },        
};

externaldef(xmptemplatewidgetclass) WidgetClass xmpTemplateWidgetClass =
			                 (WidgetClass) &xmpTemplateClassRec;


/************************************************************************
 *
 *  Initialize
 *
 ************************************************************************/
static void
Initialize (Widget request_w, Widget new_w, ArgList args, Cardinal* num_args)
{
    XmpTemplateWidget nw = (XmpTemplateWidget)new_w;

    /* Validate TemplateResource(nw) here */

    XmpGeometryInitialize(xmpTemplateWidgetClass,
				request_w, new_w, args, num_args);
}




/************************************************************************
 *
 *  Set Values
 *
 ************************************************************************/
static Boolean
SetValues (Widget old_w, Widget request_w, Widget new_w, ArgList args,
							Cardinal* num_args)
{
    XmpTemplateWidget cw = (XmpTemplateWidget)old_w;
    XmpTemplateWidget nw = (XmpTemplateWidget)new_w;
    Boolean redisplay = False;

    /* Validate TemplateResource(nw) here */

    if (TemplateResource(nw) != TemplateResource(cw)) {
	Reconfigure(nw) = True;
    }

    redisplay |= XmpGeometrySetValues(xmpTemplateWidgetClass,
					old_w, request_w, new_w, args, num_args);

    return (redisplay);
}


/************************************************************************
 *
 *  ConstraintInitialize
 *
 ************************************************************************/
static void
ConstraintInitialize (Widget rw, Widget nw, ArgList args, Cardinal* num_args)
{
    if (!XtIsRectObj (nw)) return;

    /* Validate TemplateConstraintResource(nw) here */

    XmpGeometryConstraintInitialize(xmpTemplateWidgetClass,
					rw, nw, args, num_args);
}


/************************************************************************
 *
 *  ConstraintSetValues
 *
 ************************************************************************/
static Boolean
ConstraintSetValues (Widget cw, Widget rw, Widget nw, ArgList args,
							Cardinal* num_args)
{
    XmpTemplateWidget gw = (XmpTemplateWidget)XtParent(nw);
    Boolean redisplay = False;

    if (!XtIsRectObj (nw)) return (False);

    /* Validate TemplateConstraintResource(nw) here */

    if (XtIsManaged(nw) &&
	TemplateConstraintResource(nw) != TemplateConstraintResource(cw)) {

        ConstraintReconfigure(gw) = True;
    }

    redisplay |= XmpGeometryConstraintSetValues(xmpTemplateWidgetClass,
                                cw, rw, nw, args, num_args);
 
    return(redisplay);
}



/************************************************************************
 *
 *  Size
 *
 ************************************************************************/
static void
Size (Widget w, Dimension* templateWidth, Dimension* templateHeight)
{
    /* set '*templateWidth' and '*templateHeight' to ideal size */

    /* Child list is Children(gw)[0..NumChildren(gw)-1] */

    /* Get child preferred geometry with XmpPreferredGeometry() */

    *templateWidth = 10;
    *templateHeight = 10;
}


/************************************************************************
 *
 *  Resize 
 *
 ************************************************************************/
static void
Resize (Widget w)
{
    XmpTemplateWidget gw = (XmpTemplateWidget)w;

    /* Lay out children */

    /* Set child geometry with XmpSetGeometry() */
}

/************************************************************************
 *
 *  XmpCreateTemplate
 *
 ************************************************************************/
Widget
XmpCreateTemplate(Widget parent, char* name, Arg* args, Cardinal num_args)
{
    return(XtCreateWidget(name, xmpTemplateWidgetClass, parent, args, num_args));
}

